#ifndef _CMDLINE_H
#define _CMDLINE_H

//
// Command line parser
//
// (If the network types are used, WSAStartup must have been called before calling.)
//

typedef enum tagCmdLineType
{
    cmdtApp,
    cmdtFlag,
    cmdtString,
    cmdtInt,
    cmdtFloat,
    cmdtIPV4,
    cmdtEnd
} CmdLineType;

typedef union tagCmdLineArg
{
    LPCTSTR strval;
    bool flagval;
    int intval;
    double floatval;
    unsigned long inetval;
} CmdLineArg;

typedef enum tagCmdLineFlag
{
    cmdfNormal=0,
    cmdfMandatory=1,
    cmdfPositional=2
} CmdLineFlag;

typedef struct tagCmdLineArgs
{
    LPCTSTR shortopt;
    LPCTSTR longopt;
    LPCTSTR desc;
    LPCTSTR defval;
    int flags;
    CmdLineType type;
} CmdLineArgs;

typedef struct tagCmdLineVal
{
    bool supplied;
    CmdLineArg val;
} CmdLineVal;

#define BEGIN_CMDLINE_OPTS(name) CmdLineArgs name[] = {
#define END_CMDLINE_OPTS() { NULL, NULL, NULL, NULL, 0, cmdtEnd }\
};

#define CMDLINE_APP(name, desc) { name, NULL, desc, NULL, cmdtApp },

#define CMDLINE_FLAG(shortopt, longopt, desc, flags) { shortopt, longopt, desc, NULL, flags, cmdtFlag },
#define CMDLINE_STR(shortopt, longopt, desc, flags) { shortopt, longopt, desc, NULL, flags, cmdtString },
#define CMDLINE_STR_DEF(shortopt, longopt, desc, flags, defval) { shortopt, longopt, desc, defval, flags, cmdtString },
#define CMDLINE_INT(shortopt, longopt, desc, flags) { shortopt, longopt, desc, NULL, flags, cmdtInt },
#define CMDLINE_INT_DEF(shortopt, longopt, desc, flags, defval) { shortopt, longopt, desc, defval, flags, cmdtInt },
#define CMDLINE_FLOAT(shortopt, longopt, desc, flags) { shortopt, longopt, desc, NULL, flags, cmdtFloat },
#define CMDLINE_FLOAT_DEF(shortopt, longopt, desc, flags, defval) { shortopt, longopt, desc, defval, flags, cmdtFloat },
#define CMDLINE_INET(shortopt, longopt, desc, flags) { shortopt, longopt, desc, NULL, flags, cmdtIPV4 },
#define CMDLINE_INET_DEF(shortopt, longopt, desc, flags, defval) { shortopt, longopt, desc, defval, flags, cmdtIPV4 },

class CCmdLine
{
public:
    CCmdLine(const CmdLineArgs *opts, LPCTSTR cmdline);
    CCmdLine(const CmdLineArgs *opts, int argc, char *argv[]);
    ~CCmdLine();

    bool IsValid();
    void PrintUsage();

    bool IsArgSupplied(LPCTSTR opt, bool nodefault=false);
    CmdLineArg& GetArg(LPCTSTR opt);

protected:
    LPCTSTR m_cmdline;

    int m_argc;
    char **m_argv;

    int m_numopts;
    int m_appopt;
    const CmdLineArgs *m_opts;
    CmdLineVal *m_vals;
    CmdLineVal m_noval;

    void CommonConstruct();
    void SetErrorMsg(LPCTSTR fmt, ...);
    int GetArgIndex(LPCTSTR opt, int argt=3);
    bool Parse();
    bool ParseVal(int i, LPCTSTR str);
    bool ParseCmdLine();
    bool ParseArgv();

    bool m_valid;
    LPTSTR m_errmsg;
};


#endif /* _CMDLINE_H */