/*
    smteepd: Forward a TCP connection, logging all traffic
    Copyright (C) 2001,2004 John Sullivan

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

//
// windows.h checks for _WIN32_WINNT. If this is defined and greater than 0x0400 then
// winsock2.h is included, otherwise winsock.h is included. Unless WIN32_MEAN_AND_LEAN
// is defined, in which case neither is included.
//

#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0400
#else
#if _WIN32_WINNT<0x0400
#error Need Winsock2 features available with _WIN32_WINNT>=0x0400!
#endif
#endif

#include <winsock2.h>
#include <windows.h>
#include <process.h>
#include <tchar.h>

#include <stdio.h>

#include "cmdline.h"

// Buffer sizes for network sockets

#define cs_buflen 512
#define sc_buflen 512

// A (hopefully) invalid address to start with
//unsigned long target_addr = 0x0a564378;
unsigned long target_addr = htonl(INADDR_NONE);
int target_port = -1;
int my_port = -1;
char *logpath = NULL;

FILE *opf = stdout;

// Print a text-string to a socket.
void pputs(SOCKET s, char *fmt)
{
	send(s, fmt, strlen(fmt), 0);
}

// Log a formatted string to our logfile and the console
int lprintf(char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);

    int rv = vfprintf(opf, fmt, args);
    fflush(opf);
    if (opf!=stdout) rv = vprintf(fmt, args);

    va_end(args);
    return rv;
}

// Change all awkward characters in a string to dots
void PreFilt(char *ob, int &obc)
{
    for ( int i=0 ; i<obc ; i++ ) {
        if ((ob[i]<32 || ob[i]>126) && ob[i]!='\n' && ob[i]!='\r') ob[i] = '.';
    }
}

// If we've received enough data, log it a line at a time.
void PLine(char *ib, int ibc, char *ob, int &obc, int &cnl, char *pref)
{
    memcpy(ob+obc, ib, ibc);
    PreFilt(ob+obc, ibc);
    obc += ibc;

    bool bPrint = true;

    for ( int i=0 ; i<obc ; i++ ) {
        if (ob[i]=='\n' || ob[i]=='\r' || i==256) {
            if (!i) {
                if (cnl) lprintf("%s\n", pref);
                if (ob[i]=='\n') cnl++;
            }

            if (i) lprintf("%s %.*s\n", pref, i, ob);

            int ctd = (i==256 && ob[i]!='\n' && ob[i]!='\r') ? i : i+1;

            memcpy(ob, ob+ctd, obc-ctd);
            obc -= ctd;
            i = -1;
        } else cnl = 0;
    }
}

// Per connection handler. Bi-directional pipe with logging.
unsigned __stdcall thand(void *p)
{
	SOCKET csock = (SOCKET)p;

	lprintf("Connection received\n");

    // Client-to-Server communication

	char cs_buf[cs_buflen];
    int cs_bc = 0;
    char cs_line[1024];
    int cs_lc = 0;
    int cs_nl = 0;

    // Server-to-Client communication

	char sc_buf[sc_buflen];
    int sc_bc = 0;
    char sc_line[1024];
    int sc_lc = 0;
    int sc_nl = 0;

	int rv;

    // Open the server connection

    struct sockaddr_in daddr;
    daddr.sin_family = AF_INET;
	if (target_port==-1) {
		struct servent *smtp_svc = getservbyname("smtp", NULL);
		daddr.sin_port = smtp_svc->s_port;
	} else {
		daddr.sin_port = htons(target_port);
	}
    daddr.sin_addr.s_addr = htonl(target_addr);

    SOCKET ssock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

    rv = connect(ssock, (const struct sockaddr*)&daddr, sizeof(daddr));
    if (rv==SOCKET_ERROR) {
        lprintf("Error connecting to %s:%d (%d)\n", inet_ntoa(daddr.sin_addr), ntohs(daddr.sin_port), WSAGetLastError());
        closesocket(csock);
        return 0;
    }

    struct sockaddr_in laddr;
    int laddrsz = sizeof(laddr);
    rv = getsockname(ssock, (struct sockaddr*)&laddr, &laddrsz);

	lprintf("Outgoing: %s:%d -> %s:%d\n", inet_ntoa(laddr.sin_addr), ntohs(laddr.sin_port), inet_ntoa(daddr.sin_addr), ntohs(daddr.sin_port));

    // Hook up notification handlers

    WSAEVENT hEvCli = WSACreateEvent();
    WSAEventSelect(csock, hEvCli, FD_READ|FD_WRITE|FD_CLOSE);

    WSAEVENT hEvSvr = WSACreateEvent();
    WSAEventSelect(ssock, hEvSvr, FD_READ|FD_WRITE|FD_CLOSE);

    WSAEVENT hEv[] = { hEvCli, hEvSvr };

    bool bRunning = true;

    //
    // System model and internal state flags:
    //
    //  /---\                   /-----------------v-----------------\                   /---\
    //  | C |  can read (Ccr)   |                                  |  can write (Scw)  | S |
    //  | L |------------------>| have buf (Chb)   have data (Shd) |------------------>| E |
    //  | I |                   |__________________________________|                   | R |
    //  | E |                   |                                  |                   | V |
    //  | N |<------------------| have data (Chd)  have buf (Shb)  |<------------------| E |
    //  | T |  can write (Ccw)  |                                  |  can read (Scr)   | R |
    //  \---/                   \-----------------^-----------------/                   \---/
    //
    //

    // Initial state

    bool Ccr=false, Chb=true, Shd=false, Scw=false;
    bool Ccw=false, Chd=false, Shb=true, Scr=false;

    bool Copen=true, Sopen=true; // socket hasn't been closesocket()ed
    bool Cnot=false, Snot=false; // socket has been shutdown(, SD_SEND)ed
    bool Cfin=false, Sfin=false; // have received FIN on socket

    DWORD dwTimeout = INFINITE;

    while (bRunning) {
        DWORD wv = WSAWaitForMultipleEvents(2, hEv, FALSE, dwTimeout, FALSE);

        // Handle changes in network state

        if (wv!=WSA_WAIT_TIMEOUT) {
            WSANETWORKEVENTS ne;
            WSAEnumNetworkEvents(csock, hEvCli, &ne);

            if (ne.lNetworkEvents & FD_READ) Ccr = true;
            if (ne.lNetworkEvents & FD_WRITE) Ccw = true;
            if (ne.lNetworkEvents & FD_CLOSE) {
                if (!Cfin) {
                    lprintf("Client has initiated shutdown\n");
                }
                Cfin = true;
                Ccw = false;
                // might still be unread data though
            }

            WSAEnumNetworkEvents(ssock, hEvSvr, &ne);
            if (ne.lNetworkEvents & FD_READ) Scr = true;
            if (ne.lNetworkEvents & FD_WRITE) Scw = true;
            if (ne.lNetworkEvents & FD_CLOSE) {
                if (!Sfin) {
                    lprintf("Server has initiated shutdown\n");
                }
                Sfin = true;
                Scw = false;
                // might still be unread data though
            }
        }

        // Pipe data around, if available and buffer space exists

        if (Ccr && Chb) {
            rv = recv(csock, cs_buf+cs_bc, cs_buflen-cs_bc, 0);
            if (rv==SOCKET_ERROR) {
                rv = WSAGetLastError();
                if (rv==WSAEWOULDBLOCK) {
                    Ccr = false;
                } else {
                    lprintf("Client socket read error %d\n", rv);
                }
            } else {
                PLine(cs_buf+cs_bc, rv, cs_line, cs_lc, cs_nl, ">>>");

                cs_bc += rv;

                if (cs_bc>0) Shd = true;
                if (cs_bc>=cs_buflen) Chb = false;
            }
        }

        if (Ccw && Chd) {
            rv = send(csock, sc_buf, sc_bc, 0);
            if (rv==SOCKET_ERROR) {
                rv = WSAGetLastError();
                if (rv==WSAEWOULDBLOCK) {
                    Ccw = false;
                } else {
                    lprintf("Client socket send error %d\n", rv);
                }
            } else {
                if (rv<sc_bc) memcpy(sc_buf, sc_buf+rv, sc_bc-rv);
                sc_bc -= rv;

                if (!sc_bc) Chd = false;
                if (sc_bc<sc_buflen) Shb = true;
            }
        }

        if (Scr && Shb) {
            rv = recv(ssock, sc_buf+sc_bc, sc_buflen-cs_bc, 0);
            if (rv==SOCKET_ERROR) {
                rv = WSAGetLastError();
                if (rv==WSAEWOULDBLOCK) {
                    Scr = false;
                } else {
                    lprintf("Server socket read error %d\n", rv);
                }
            } else {
                PLine(sc_buf+sc_bc, rv, sc_line, sc_lc, sc_nl, "<<<");

                sc_bc += rv;

                if (sc_bc) Chd = true;
                if (sc_bc>=sc_buflen) Shb = false;
            }
        }

        if (Scw && Shd) {
            rv = send(ssock, cs_buf, cs_bc, 0);
            if (rv==SOCKET_ERROR) {
                rv = WSAGetLastError();
                if (rv==WSAEWOULDBLOCK) {
                    Scw = false;
                } else {
                    lprintf("Server socket send error %d\n", rv);
                }
            } else {
                if (rv<cs_bc) memcpy(cs_buf, cs_buf+rv, cs_bc-rv);
                cs_bc -= rv;

                if (!cs_bc) Shd = false;
                if (cs_bc<cs_buflen) Chb = true;
            }
        }

        // Handle various shades of being closed

        if (Cfin && !Ccr && Copen) {
            WSAEventSelect(csock, hEvCli, 0);
            closesocket(csock);
            Copen = false;
        }
        if (Cfin && !cs_bc && Sopen && !Snot) {
            shutdown(ssock, SD_SEND);
            Snot = true;
        }

        if (Sfin && !Scr && Sopen) {
            WSAEventSelect(ssock, hEvSvr, 0);
            closesocket(ssock);
            Sopen = false;
        }
        if (Sfin && !sc_bc && Copen && !Cnot) {
            shutdown(csock, SD_SEND);
            Cnot = true;
        }

        // Check to see whether there is still work outstanding

        dwTimeout = ((Ccr && Chb) || (Shd && Scw) || (Ccw && Chd) || (Shb && Scr)) ? 0 : INFINITE;

        // Terminate on receiving 2 FINs, not on issuing two closesocket()s.
        // (Both ends could exit simultaneously, leaving unread data
        // in the pipe.)

        if (Cfin && Sfin) bRunning = false;
    }

    WSACloseEvent(hEvCli);
    WSACloseEvent(hEvSvr);

    if (cs_bc || sc_bc) {
        lprintf("Connection closed (unsent data! C->S: %d , S->C: %d\n", cs_bc, sc_bc);
    } else {
        lprintf("Connection closed\n");
    }

	return 0;
}

// Main listening loop. No quit option, just close the Win32 console window!

//
// Command-line/usage:
//
// smteepd [-f logfile] [-t] target_addr [-p] target_port [-l] listen_port
//

BEGIN_CMDLINE_OPTS(smteepd_opts)
    CMDLINE_APP(_T("smteepd"), _T("TCP connection logger"))

    CMDLINE_INET(_T("t"), _T("target"), _T("remote IP address to connect to"), cmdfMandatory|cmdfPositional)
    CMDLINE_INT_DEF(_T("p"), _T("target-port"), _T("remote port to connect to"), cmdfPositional, _T("25"))
    CMDLINE_INT_DEF(_T("l"), _T("listen-port"), _T("local port to listen on"), cmdfPositional, _T("25"))
    CMDLINE_STR_DEF(_T("f"), _T("logfile"), _T("path of output log file"), cmdfNormal, _T("smteepd.log"))

END_CMDLINE_OPTS()

int main(int argc, char *argv[])
{
    CCmdLine cmdLine(smteepd_opts, argc, argv);

    if (!cmdLine.IsValid()) {
        cmdLine.PrintUsage();
        return 1;
    }

    target_addr = cmdLine.GetArg(_T("t")).inetval;
    target_port = cmdLine.GetArg(_T("p")).intval;
    my_port = cmdLine.GetArg(_T("l")).intval;
    logpath = strdup(cmdLine.GetArg(_T("f")).strval);

    if (target_addr==INADDR_NONE) {
        fprintf(stderr, "Invalid target address\n");
        return 1;
    }
    if (target_addr==INADDR_LOOPBACK && my_port==target_port) {
        fprintf(stderr, "Can't listen on target socket\n");
        return 1;
    }

	WSADATA wsadata;
	WSAStartup(MAKEWORD(2, 0), &wsadata);

	SOCKET lsock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

	int we = WSAGetLastError();

	struct sockaddr_in saddr;
	saddr.sin_family = AF_INET;
    if (my_port==-1) {
	    if (target_port==-1) {
	        struct servent *smtp_svc = getservbyname("smtp", NULL);
	        saddr.sin_port = smtp_svc->s_port;
	    } else {
		    saddr.sin_port = htons(target_port);
	    }
    } else {
        saddr.sin_port = htons(my_port);
    }

	saddr.sin_addr.s_addr = htonl(INADDR_ANY);

	int rv = bind(lsock, (const struct sockaddr*)&saddr, sizeof(saddr));

	rv = listen(lsock, SOMAXCONN);

	struct sockaddr_in bsaddr;
	int bsz;
	getsockname(lsock, (struct sockaddr*)&bsaddr, &bsz);

	BOOL sso = TRUE;
	setsockopt(lsock, SOL_SOCKET, SO_BROADCAST, (char*)&sso, sizeof(sso));

    if (strcmp(logpath, "-")!=0) opf = fopen(logpath, "w");
    lprintf("Listening...\n");

	while (1) {

		struct sockaddr_in aaddr;
		int asz = sizeof(aaddr);
		aaddr.sin_family = AF_INET;
		SOCKET nsock = accept(lsock, (struct sockaddr*)&aaddr, &asz);
		we = WSAGetLastError();

        lprintf("Incoming: %s:%d -> %s:%d\n", inet_ntoa(aaddr.sin_addr), ntohs(aaddr.sin_port), inet_ntoa(saddr.sin_addr), ntohs(saddr.sin_port));

		unsigned threadid;
		HANDLE hT = (HANDLE)_beginthreadex(NULL, 0, thand, (void*)nsock, 0, &threadid);
		CloseHandle(hT);
	}

	return 0;
}
